"""
Create a probability table 

    pdf_table(d::UnivariateDistribution, eps::Float, tol::Float64)

Given a distribution `d`, the interval `[0,quantile(d, tol)]` is 
divided in `n` sub-intervals of length `eps`. The probabilities 
P_n for `x` falling in each subinterval are returned as a vector.

## Arguments
- d: An `UnivariateDistribution`. See `Distributions.jl`.
- eps: Discretization step size.
- tol: The quantile that marks the end of the interval.

## Examples
```julia-repl
julia> pdf_t = pdf_table(Normal(10.0, 0.1), 0.01, 0.9999)
```
"""
function pdf_table(d::Distributions.UnivariateDistribution, eps::Float64, tol::Float64)

    n::Int64 = round(Int64, Distributions.quantile(d, tol)/eps)

    pdf = zeros(Float64, n)
    cd1 = 0.0
    for i in 1:n-1
        cd2 = Distributions.cdf(d, i*eps)
        pdf[i] = cd2-cd1
        cd1 = cd2
    end
    pdf[end] = 1.0-sum(pdf)

    return pdf
end


"""
Solve uSEIR: arbitrary distributions for infected anf exposed times

    solve_seir(de, di, eps, rate, dtr = 1, i0 = 1.0E-6, stp = 1.0E-6)

Given the distributions for the time that people are exposed and infected, this 
routines solver the SEIR model.

## Arguments
- de, di: `UnivariateDistribution`. The distributions for the time that people 
  are exposed and infected respectively. If a `Float64` is introduced the program 
  interpret this as the distribution being a delta function.
- eps: `Float64`. The discretization parameter.
- rate: `Float64`. The infection rate.
- dtr: `Int64`. Optional (default 1). How often one wants the values of 
  `s(t), e(t), i(t), r(t)` be returned
- i0: `Float64`. Optional (default 1.0E-6). Ratio of people infected at `t=0`.
- stp: `Float64`. Optional (default 1.0E-10). Stoping criteria: Simulation ends when 
  e(t) and i(t) are both smaller than `stp`.

## Returns
- solution: `Vector{Float64}`. The values of `s(t), e(t), i(t), r(t)` for each time. 

## Examples
```julia-repl
julia> slv = solve_uSEIR(Gamma(5.0, 1.0), Gamma(7.0, 1.4), 0.01, 2.0/7.0)
julia> # Get the values of the infected
julia> infected = slv(3:4:end)
julia> # Get the values of the exposed
julia> exposed  = slv(2:4:end)
julia> # Get the values of the recovered
julia> inf = slv(4:4:end)
```
"""
function solve_uSEIR(de::Distributions.UnivariateDistribution, di::Distributions.UnivariateDistribution, eps::Float64, prob::Float64, dtr::Int64=1, i0::Float64 = 1.0E-6, stp::Float64 = 1.0E-10)
    
    # Build probability table
    pn  = prob*eps
    pdi = pdf_table(di, eps, 0.9999)
    ni  = length(pdi)
    pde = pdf_table(de, eps, 0.9999)
    ne  = length(pde)

    # set initial condition
    S    = 1.0-i0
    E    = zeros(ne)
    E[1] = i0
    I    = zeros(ni)
    R    = 0.0

    sol = Vector{Float64}()

    n  = 0
    sI = 0.0
    while true
        n = n + 1
        R = R + I[1]
        for k in 1:ni-1
	    I[k] = I[k+1] + pdi[k]*E[1]
        end
        I[end] = pdi[end]*E[1]

        for k in 1:ne-1
	    E[k] = E[k+1] + pn*pde[k]*sI*S
        end
        E[end] = pn*pde[end]*sI*S

        S = S - pn*sI*S

        sI = sum(I)
        sE = sum(E)
        
        if (mod(n-1,dtr) == 0)
            append!(sol, [S, sE, sI, R])
        end

        if ( (sE < stp) && (sI < stp) ) 
	    break
        end

    end
    
    return sol
end

function solve_uSEIR(te::Float64, ti::Float64, eps::Float64, prob::Float64, dtr::Int64=1, i0::Float64 = 1.0E-6, stp::Float64 = 1.0E-10)
    
    pn = prob*eps
    ni = round(Int64, ti/eps)
    ne = round(Int64, te/eps)
    
    # set initial condition
    S = 1.0-i0
    E = zeros(ne)
    E[1] = i0
    I = zeros(ni)
    R = 0.0

    sol = Vector{Float64}()

    n  = 0
    sI = 0.0
    while true
        n = n + 1
        R = R + I[1]
        for k in 1:ni-1
	    I[k] = I[k+1]
        end
        I[end] = E[1]

        for k in 1:ne-1
	    E[k] = E[k+1]
        end
        E[end] = pn*sI*S

        S = S - pn*sI*S

        sI = sum(I)
        sE = sum(E)
        
        if (mod(n-1,dtr) == 0)
            append!(sol, [S, sE, sI, R])
        end

        if ( (sE < stp) && (sI < stp) ) 
	    break
        end

    end
    
    return sol
end

function solve_uSEIR(de::Distributions.UnivariateDistribution, ti::Float64, eps::Float64, prob::Float64, dtr::Int64=1, i0::Float64 = 1.0E-6, stp::Float64 = 1.0E-10)
    
    pn  = prob*eps
    ni  = round(Int64, ti/eps)
    pde = pdf_table(de, eps, 0.9999)
    ne  = length(pde)
    
    # set initial condition
    S = 1.0-i0
    E = zeros(ne)
    E[1] = i0
    I = zeros(ni)
    R = 0.0

    sol = Vector{Float64}()

    n  = 0
    sI = 0.0
    while true
        n = n + 1
        R = R + I[1]
        for k in 1:ni-1
	    I[k] = I[k+1]
        end
        I[end] = E[1]

        for k in 1:ne-1
	    E[k] = E[k+1] + pn*pde[k]*sI*S
        end
        E[end] = pn*pde[end]*sI*S

        S = S - pn*sI*S

        sI = sum(I)
        sE = sum(E)
        
        if (mod(n-1,dtr) == 0)
            append!(sol, [S, sE, sI, R])
        end

        if ( (sE < stp) && (sI < stp) ) 
	    break
        end

    end
    
    return sol
end

function solve_uSEIR(te::Float64, di::Distributions.UnivariateDistribution, eps::Float64, prob::Float64, dtr::Int64=1, i0::Float64 = 1.0E-6, stp::Float64 = 1.0E-10)
    
    pn = prob*eps
    pdi = pdf_table(di, eps, 0.9999)
    ni  = length(pdi)
    ne = round(Int64, te/eps)
    
    # set initial condition
    S = 1.0-i0
    E = zeros(ne)
    E[1] = i0
    I = zeros(ni)
    R = 0.0

    sol = Vector{Float64}()

    n  = 0
    sI = 0.0
    while true
        n = n + 1
        R = R + I[1]
        for k in 1:ni-1
	    I[k] = I[k+1] + pdi[k]*E[1]
        end
        I[end] = pdi[end]*E[1]

        for k in 1:ne-1
	    E[k] = E[k+1]
        end
        E[end] = pn*sI*S

        S = S - pn*sI*S

        sI = sum(I)
        sE = sum(E)
        
        if (mod(n-1,dtr) == 0)
            append!(sol, [S, sE, sI, R])
        end

        if ( (sE < stp) && (sI < stp) ) 
	    break
        end

    end
    
    return sol
end
